#include <tcl.h>
#include <tclInt.h>
#include <tclIntDecls.h>
#include <Judy.h>

#ifndef max
#define max(a,b) ((a) >= (b) ? (a) : (b))
#endif

static void DupJudyInternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr);
static void FreeJudyInternalRep(Tcl_Obj *judyPtr);
static void UpdateStringOfJudy(Tcl_Obj *judyPtr);
static int SetJudyFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr);

/*
 * Internal representation of a judy tree.
 *
 * The internal representation of a judy tree object is a hash table
 * (with Tcl_Objs for both keys and values), a reference count and
 * epoch number for detecting concurrent modifications of the
 * judy tree, and a pointer to the parent object (used when
 * invalidating string reps of pathed judy tree trees) which is NULL
 * in normal use.  The fact that hash tables know (with appropriate
 * initialisation) already about objects makes key management /so/
 * much easier!
 *
 * Reference counts are used to enable safe iteration across hashes
 * while allowing the type of the containing object to be modified.
 */

typedef struct Judy {
    void *table;	/* JudySL array. */
    int epoch;
    int refcount;
    struct Judy *chain;
    int maxkey;	/* maximum key size */
    int count;	/* content count */
} Judy;

/*
 * The structure below defines the judy tree object type by means of
 * procedures that can be invoked by generic object code.
 */

Tcl_ObjType JudyType = {
    "judy",
    FreeJudyInternalRep,		/* freeIntRepProc */
    DupJudyInternalRep,		        /* dupIntRepProc */
    UpdateStringOfJudy,			/* updateStringProc */
    SetJudyFromAny			/* setFromAnyProc */
};

/*
 *----------------------------------------------------------------------
 *
 * DupJudyInternalRep --
 *
 *	Initialize the internal representation of a judy tree Tcl_Obj
 *	to a copy of the internal representation of an existing
 *	judy tree object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"srcPtr"s judy tree internal rep pointer should not be NULL and
 *	we assume it is not NULL. We set "copyPtr"s internal rep to a
 *	pointer to a newly allocated judy tree rep that, in turn, points
 *	to "srcPtr"s key and value objects. Those objects are not
 *	actually copied but are shared between "srcPtr" and "copyPtr".
 *	The ref count of each key and value object is incremented.
 *
 *----------------------------------------------------------------------
 */

static void DupJudyInternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr)
{
    Judy *oldJudy = (Judy *) srcPtr->internalRep.otherValuePtr;
    Judy *newJudy = (Judy *)ckalloc(sizeof(Judy));
    Tcl_Obj **value;
    PJError_t PJError;
    char *key = ckalloc(oldJudy->maxkey);

    //fprintf(stderr, "Dup Judy: %04x\n", srcPtr);

    /*
     * Copy values across from the old hash table.
     */
    for (key[0] = '\0', value = (Tcl_Obj **)JudySLFirst(oldJudy->table, key, PJError);
	 value != NULL;
	 value = (Tcl_Obj **)JudySLNext(oldJudy->table, key, PJError)) {
	Tcl_Obj **vptr = (Tcl_Obj **)JudySLIns(&newJudy->table, key, PJError);
	if (vptr) {
	    *vptr = *value;
	    Tcl_IncrRefCount(*value);
	} else {
	    /* Error - don't know what we do about that */
	}
    }

    /*
     * Initialise other fields.
     */
    newJudy->epoch = 0;
    newJudy->chain = NULL;
    newJudy->refcount = 1;
    newJudy->maxkey = oldJudy->maxkey;
    newJudy->count = oldJudy->count;

    /*
     * Store in the object.
     */
    copyPtr->internalRep.otherValuePtr = (VOID *) newJudy;
    copyPtr->typePtr = &JudyType;
    ckfree(key);
}

/*
 *----------------------------------------------------------------------
 *
 * FreeJudyInternalRep --
 *
 *	Deallocate the storage associated with a judy tree object's
 *	internal representation.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Frees the memory holding the judy tree's internal hash table.
 *	Decrements the reference count of all key and value objects,
 *	which may free them.
 *
 *----------------------------------------------------------------------
 */

static void FreeJudyInternalRep(Tcl_Obj *judyPtr)
{
    Judy *judy = (Judy *) judyPtr->internalRep.otherValuePtr;
    PJError_t PJError;

    //fprintf(stderr, "Free Judy: %04x\n", judyPtr);

    --judy->refcount;
    if (judy->refcount <= 0) {
	JudySLFreeArray(&judy->table, PJError);
    }

    judyPtr->internalRep.otherValuePtr = NULL;	/* Belt and braces! */
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfJudy --
 *
 *	Update the string representation for a judy tree object.
 *	Note: This procedure does not invalidate an existing old string
 *	rep so storage will be lost if this has not already been done.
 *	This code is based on UpdateStringOfList in tclListObj.c
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The object's string is set to a valid string that results from
 *	the judy-to-string conversion. This string will be empty if the
 *	judy tree has no key/value pairs. The judy tree internal
 *	representation should not be NULL and we assume it is not NULL.
 *
 *----------------------------------------------------------------------
 */

static void UpdateStringOfJudy(Tcl_Obj *judyPtr)
{
    PJError_t PJError;
#define LOCAL_SIZE 20
    int localFlags[LOCAL_SIZE], *flagPtr;
    Judy *judy = (Judy *) judyPtr->internalRep.otherValuePtr;
    int numElems, i, length;
    char *elem, *dst;
    char *key = ckalloc(judy->maxkey);
    Tcl_Obj **value;

    //fprintf(stderr, "Update Judy: %04x\n", judyPtr);

    /*
     * Pass 0: get number of elements
     */

    numElems = judy->count;
    if (numElems <= LOCAL_SIZE) {
	flagPtr = localFlags;
    } else {
	flagPtr = (int *) ckalloc((unsigned) numElems*sizeof(int));
    }

    //fprintf(stderr, "Update Judy 1: %d\n", numElems);

    /*
     * Pass 1: estimate space, gather flags.
     */

    judyPtr->length = 1;
    for (key[0] = '\0', i=0, value = (Tcl_Obj **)JudySLFirst(judy->table, key, PJError);
	 value != NULL;
	 value = (Tcl_Obj **)JudySLNext(judy->table, key, PJError), i++) {
	//fprintf(stderr, "Judy key: %s\n", key);
	judyPtr->length += strlen(key) + 1;
	elem = Tcl_GetStringFromObj(*value, &length);
	//fprintf(stderr, "Judy elem: %s\n", elem);
	judyPtr->length += Tcl_ScanCountedElement(elem, length, &flagPtr[i]) + 1;
	//fprintf(stderr, "Judy len: %d\n", judyPtr->length);
    }

    /*
     * Pass 2: copy into string rep buffer.
     */

    //fprintf(stderr, "Update Judy 2: %d\n", judyPtr->length);

    judyPtr->bytes = ckalloc((unsigned) judyPtr->length);
    dst = judyPtr->bytes;
    for (i=0, key[0] = '\0', value = (Tcl_Obj **)JudySLFirst(judy->table, key, PJError);
	 i < numElems;
	 value = (Tcl_Obj **)JudySLNext(judy->table, key, PJError), i++) {
	//fprintf(stderr, "Judy key: %s\n", key);
	strcpy(dst, key);
	dst += strlen(key);
	*(dst++) = ' ';

	elem = Tcl_GetStringFromObj(*value, &length);
	//fprintf(stderr, "Judy elem: %s\n", elem);
	dst += Tcl_ConvertCountedElement(elem, length, dst, flagPtr[i+1]);
	*(dst++) = ' ';
	*(dst+1) = '\0';
	//fprintf(stderr, "Judy to string: %04x\n", dst);
    }
    if (flagPtr != localFlags) {
	ckfree((char *) flagPtr);
    }
    if (dst == judyPtr->bytes) {
	*dst = 0;
    } else {
	*(--dst) = 0;
    }
    judyPtr->length = dst - judyPtr->bytes;
    ckfree(key);
    //fprintf(stderr, "Update Judy done: %04x\n", judyPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * SetJudyFromAny --
 *
 *	Convert a non-judy tree object into a judy tree object.  This
 *	code is very closely related to SetListFromAny in tclListObj.c
 *	but does not actually guarantee that a judy tree object will
 *	have a string rep (as conversions from lists are handled with a
 *	special case.)
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	If the string can be converted, it loses any old internal
 *	representation that it had and gains a judy tree's internalRep.
 *
 *----------------------------------------------------------------------
 */

static int SetJudyFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr)
{
    Tcl_ObjType *oldTypePtr = objPtr->typePtr;

    //fprintf(stderr, "Set Judy: %04x\n", objPtr);

    /*
     * Since lists and judy trees have very closely-related string
     * representations (i.e. the same parsing code) we can safely
     * special-case the conversion from lists to judy trees.
     */

    if (oldTypePtr == &tclListType) {
	register Tcl_Obj **value;
	register char *key;
	Tcl_Obj **objv;
	Judy *judy;
	PJError_t PJError;
	int objc, i;

	if (Tcl_ListObjGetElements(interp, objPtr, &objc, &objv) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (objc & 1) {
	    if (interp != NULL) {
		Tcl_SetObjResult(interp,
			Tcl_NewStringObj("missing value to go with key", -1));
	    }
	    return TCL_ERROR;
	}

	/* create a new Judy object */
	judy = (Judy *) ckalloc(sizeof(Judy));
	judy->table = NULL;
	judy->maxkey = 0;
	judy->count = 0;
	judy->epoch = 0;
	judy->chain = NULL;
	judy->refcount = 1;
	
	/*
	 * Build the hash of key/value pairs.
	 */
	for (i=0 ; i<objc ; i+=2) {
	    /*
	     * Store key and value in the hash table we're building.
	     */
	    key = Tcl_GetString(objv[i]);
	    value = (Tcl_Obj **)JudySLGet(judy->table, key, PJError);
	    if (value) {
		/* delete old duplicate from judy table */
		Tcl_DecrRefCount(*value);
	    } else {
		value = (Tcl_Obj **)JudySLIns(&judy->table, key, PJError);
		judy->count++;
	    }
	    *value = objv[i+1];
	    judy->maxkey = max(judy->maxkey, strlen(key));
	    Tcl_DecrRefCount(objv[i]);	/* we use the key's string rep only */
	}

	/*
	 * Free the old internalRep before setting the new one. We do this as
	 * late as possible to allow the conversion code, in particular
	 * Tcl_GetStringFromObj, to use that old internalRep.
	 */

	if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	    oldTypePtr->freeIntRepProc(objPtr);
	}

	objPtr->internalRep.otherValuePtr = (VOID *) judy;
	objPtr->typePtr = &JudyType;

	return TCL_OK;
    } else {
	/* cheat - make this a List first, then recurse */
	if (tclListType.setFromAnyProc(interp, objPtr) == TCL_ERROR)
	    return TCL_ERROR;
	else
	    return SetJudyFromAny(interp, objPtr);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewDictObj --
 *
 *	This procedure is normally called when not debugging: i.e., when
 *	TCL_MEM_DEBUG is not defined. It creates a new dict object
 *	without any content.
 *
 *	When TCL_MEM_DEBUG is defined, this procedure just returns the
 *	result of calling the debugging version Tcl_DbNewDictObj.
 *
 * Results:
 *	A new dict object is returned; it has no keys defined in it.
 *	The new object's string representation is left NULL, and the
 *	ref count of the object is 0.
 *
 * Side Effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Obj *Tcl_NewJudyObj()
{
#ifdef TCL_MEM_DEBUG
    return Tcl_DbNewJudyObj("unknown", 0);
#else /* !TCL_MEM_DEBUG */
    Tcl_Obj *judyPtr;
    Judy *judy;

    TclNewObj(judyPtr);
    //fprintf(stderr, "JudyPtr: %04x\n", judyPtr);
    Tcl_InvalidateStringRep(judyPtr);
    judy = (Judy *) ckalloc(sizeof(Judy));
    //fprintf(stderr, "judy: %04x\n", judy);
    judy->table = NULL;
    judy->epoch = 0;
    judy->refcount = 1;
    judy->chain = NULL;
    judy->maxkey = 0;
    judy->count = 0;
    judyPtr->internalRep.otherValuePtr = (VOID *) judy;
    judyPtr->typePtr = &JudyType;
    //fprintf(stderr, "done judy: %04x\n", judy);
    return judyPtr;
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_DbNewJudyObj --
 *
 *	This procedure is normally called when debugging: i.e., when
 *	TCL_MEM_DEBUG is defined. It creates new judy objects. It is the
 *	same as the Tcl_NewJudyObj procedure above except that it calls
 *	Tcl_DbCkalloc directly with the file name and line number from its
 *	caller. This simplifies debugging since then the [memory active]
 *	command	will report the correct file name and line number when
 *	reporting objects that haven't been freed.
 *
 *	When TCL_MEM_DEBUG is not defined, this procedure just returns the
 *	result of calling Tcl_NewJudyObj.
 *
 * Results:
 *	A new judy object is returned; it has no keys defined in it.
 *	The new object's string representation is left NULL, and the
 *	ref count of the object is 0.
 *
 * Side Effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Obj *Tcl_DbNewJudyObj(const char *file, int line)
{
#ifdef TCL_MEM_DEBUG
    Tcl_Obj *judyPtr;
    Judy *judy;

    TclDbNewObj(judyPtr, file, line);
    Tcl_InvalidateStringRep(judyPtr);
    judy = (Judy *) ckalloc(sizeof(Judy));
    judy->table = NULL;
    judy->epoch = 0;
    judy->refcount = 1;
    judy->chain = NULL;
    judy->maxkey = 0;
    judy->count = 0;
    judyPtr->internalRep.otherValuePtr = (VOID *) judy;
    judyPtr->typePtr = &JudyType;
    return judyPtr;
#else /* !TCL_MEM_DEBUG */
    return Tcl_NewJudyObj();
#endif
}

